"""Template for the raiconfig.yml file generated by the init command."""

CONFIG_TEMPLATE = """
active_profile: dev

profiles:
  dev:
    # Default connection for this profile
    default_connection: snowflake_dev

    # Connections available in this profile
    connections:
      snowflake_dev:
        type: snowflake
        authenticator: username_password
        account: ""
        user: ""
        password: ""
        warehouse: ""
        database: ""
        schema_: ""
        role: ""

      duckdb_local:
        type: duckdb
        path: ":memory:"

  prod:
    # Default connection for this profile
    default_connection: snowflake_prod

    # Connections available in this profile
    connections:
      snowflake_prod:
        type: snowflake
        authenticator: username_password
        account: ""
        user: ""
        password: ""
        warehouse: ""
        database: ""
        schema_: ""
        role: ""
"""

