# In your_package/submodule.py
from .internal import Relationship

external = Relationship.builtins["external"]
concept_population = Relationship.builtins["concept_population"]
function = Relationship.builtins["function"]
from_cdc = Relationship.builtins["from_cdc"]
track = Relationship.builtins["track"]
recursion_config = Relationship.builtins["recursion_config"]
