"""
Snowflake resources module.
"""
# Import order matters - Resources must be imported first since other classes depend on it
from .snowflake import Resources, Provider, Graph, SnowflakeClient, APP_NAME, PYREL_ROOT_DB, ExecContext, INTERNAL_ENGINE_SIZES, ENGINE_SIZES_AWS, ENGINE_SIZES_AZURE, PrimaryKey

# These imports depend on Resources, so they come after
from .cli_resources import CLIResources
from .use_index_resources import UseIndexResources
from .direct_access_resources import DirectAccessResources
from .resources_factory import create_resources_instance

__all__ = [
    'Resources', 'DirectAccessResources', 'Provider', 'Graph', 'SnowflakeClient',
    'APP_NAME', 'PYREL_ROOT_DB', 'CLIResources', 'UseIndexResources', 'ExecContext',
    'INTERNAL_ENGINE_SIZES', 'ENGINE_SIZES_AWS', 'ENGINE_SIZES_AZURE', 'PrimaryKey',
    'create_resources_instance',
]


